	PROGRAM heat

! ============================================================
!
!  Exercise 7 - a heat transfer example - parallel version
!  
! ============================================================
!
!  declarations
!
! ============================================================

	IMPLICIT NONE

	INTEGER, PARAMETER :: iMax = 81, kMax = 81
        INTEGER, PARAMETER :: itMax = 20000

        INTEGER :: it, i, k

        DOUBLE PRECISION :: eps = 1.d-08


        DOUBLE PRECISION, DIMENSION(iMax, kMax) :: phi, phin

!HPF$	DISTRIBUTE (BLOCK, BLOCK) :: phi, phin
!HPF$	SHADOW (1:1, 1:1) :: phi, phin

	DOUBLE PRECISION :: dx, dy, dx2, dy2
        DOUBLE PRECISION :: dx2i, dy2i, dt, dPhi, dPhiMax


        INTEGER startTime, stopTime, cRate

! ============================================================
!
!  main program
!
! ============================================================


!	initialize constants
! ---------------------------------
	dx = 1.d0/iMax
	dy = 1.d0/kMax
	dx2 = dx*dx
	dy2 = dy*dy

	dx2i = 1.d0/dx2
	dy2i = 1.d0/dy2

	dt = MIN(dx2, dy2)/4.d0


!	initialize array
! ---------------------------------
	phi = 0.d0

	phi(:,kMax)=1.d0

	DO k=2,kMax-1
	  phi(1,k)=phi(1,k-1)+dy
	  phi(iMax,k)=phi(iMax,k-1)+dy
	END DO

	PRINT *, "start calculating..."

!	time measuring...
!----------------------------------
	CALL SYSTEM_CLOCK(count_rate=cRate)
	CALL SYSTEM_CLOCK(count=startTime)


!	iteration
! ---------------------------------
	phin = phi

	DO it=1, itMax
	  dPhiMax=0.d0

!HPF$ INDEPENDENT, REDUCTION(dPhiMax)
	  DO i=2,iMax-1
!HPF$ INDEPENDENT, REDUCTION(dPhiMax), ON HOME(phi(i,k)), NEW(dPhi)
	    DO k=2,kMax-1

              dPhi = (phi(i+1,k) + phi(i-1,k) - 2.d0*phi(i,k))*dx2i
	      dPhi = dPhi + (phi(i,k+1) + phi(i,k-1) - 2.d0*phi(i,k))*dy2i

	      dPhi = dPhi*dt

	      dPhiMax = MAX(dPhiMax, dPhi)
	      phin(i,k) = phi(i,k)+dPhi

	    END DO
	  END DO

	  phi = phin

	  IF (dPhiMax .LT. eps) THEN
            EXIT
          END IF

	END DO

!	print stats
!----------------------------------

	CALL SYSTEM_CLOCK(count=stopTime)

	PRINT *, "iterations:",it
	PRINT *, "time [ms] :", 1000*(stopTime-startTime)/cRate

	END PROGRAM heat
